/***************************************************************\
|                  I M P L E M E N T A T I O N                  |
\***************************************************************/

#include "pg.h"
#include "bmp.c"

/***************************************************************\
|                Compression algorithm handling                 |
\***************************************************************/

//Reserves a memory space to allocate an image
unsigned char* getMemory(int size) {
	return (unsigned char*)malloc(sizeof(unsigned char*)*size);
}

//Puts the first pixel in the output array
void put1(unsigned char pix,unsigned char* output,int addr) {
	output[addr]=(output[addr]&0x0F)+(pix&0xF0);
}
//Puts the second pixel in the output array
void put2(unsigned char pix, unsigned char* output,int addr) {
	output[addr]=(output[addr]&0xF0)+(pix&0x0F);
}

//Determines where the transposed byte must be saved
int transpose(int x,int n,int m) {
	n=(n+7)&0xF8;
	return ((x)%m)*(n)+2*(int)(x/m);
}

//B3 and B4 expansion algorithm sub function
unsigned char popBit(unsigned char *byte) {
  unsigned char bit=(*byte)&1;
  (*byte)>>=1;
  return bit;
}

//Expands B3/B4 algorithm
void expandB3B4(char* array,tImage* img, int *i,int cursor) {
	int pos;
	unsigned char maskbyte,rep;
	for(pos=0;pos<0x400;(*img).pix[pos++]=0); //clean output garbage
  while ((cursor*2)<((*img).size)) {
	  maskbyte=array[(*i)++];
	  for (char k=8;k&&(cursor*2)<((*img).size);k--) {
	    if (popBit(&maskbyte)) {
	      (*img).pix[cursor++]=array[(*i)++];
	    } else {
	      pos=66+((0x100)*((rep=array[(*i)++])&3))+(unsigned char)array[(*i)++];
	      rep=(rep>>2)+3;
	      while (rep--) (*img).pix[(cursor++)%(0x400)]=(*img).pix[(pos++)%(0x400)];
	    }
	  }
	}
}

//Expands an array into an image
int mExpandGraphic(char* array,tImage *image, int i) {
	/*
		Reads array and extracts tImage
		returns the next image address or -1 in case of error

		Header info:
		 char checksum, short int height, short int width, (char)0, char compressionType
		 normaly:
		 checksum - height - 00 - width - 00 - 00 - compression type
	*/

  int cursor=0;

  //Get memory for the image
	(*image).height=array[++i];
	(*image).width =array[i+=2];
	(*image).size  =(*image).height*(*image).width;
	i+=3;

  switch ((unsigned char)array[i++]) {
		case PG_COMP_RAW: //No Compression Algorithm
		  if (((*image).pix=getMemory((*image).size))==NULL) return -1;
		  while ((cursor*2)<(*image).size) {
				(*image).pix[cursor++]=array[i++];
			}
			break;
		case PG_COMP_RLE_LR: //RLE Left to Right Compression Algorithm
		  if (((*image).pix=getMemory((*image).size))==NULL) return -1;
		  while ((cursor*2)<(*image).size) {
				if (array[i]<0) {
					//negative
					while (array[i]++) {
						(*image).pix[cursor++]=array[i+1];
					}
					i+=2;
				} else {
					//Positive
					char cx=array[i]+1;
					while (cx--) {
						(*image).pix[cursor++]=array[++i];
					}
					i++;
				}
			}
			break;
		case PG_COMP_RLE_UD: //RLE Up to Down Compression Algorithm
		  if (((*image).pix=getMemory((*image).size))==NULL) return -1;
			{int t; //I use the { to declare local t variable
		  while ((cursor*2)<(*image).size) {
				if (array[i]<0) {
					//negative
					while (array[i]++) {
						t=transpose(cursor++,

						(*image).width+3-(((*image).width-1)&3),

						(*image).height);
						put1(array[i+1],(*image).pix,t>>1);
						//If there is a wasted pix don't show it
						if ((t+1)%(*image).width) put2(array[i+1],(*image).pix,t>>1);
					}
					i+=2;
				} else {
					//Positive
					char cx=array[i]+1;
					while (cx--) {
						t=transpose(cursor++,

						(*image).width+3-(((*image).width-1)&3),

						(*image).height);
						put1(array[++i],(*image).pix,t>>1);
						//If there is a wasted pix don't show it
						if ((t+1)%(*image).width) put2(array[i],(*image).pix,t>>1);
					}
					i++;
				}
			}}
			break;
		case PG_COMP_LZX_LR: //LZ Groody Up to Down Version Compression Algorithm
		  if (((*image).pix=getMemory(0x400))==NULL) return -1;
			expandB3B4(array,image,&i,cursor);
//			(*image).size+=(*image).height*(((*image).width+7)&7);
			break;
		case PG_COMP_LZX_UD: //LZ Groody Left to Right Version Compression Algorithm
		  if (((*image).pix=getMemory(0x400))==NULL) return -1;
//			(*image).size+=(*image).height*(((*image).width+7)&7);

			{unsigned char* outputaux=getMemory((*image).size);
			expandB3B4(array,image,&i,cursor);
			//Transpose
			int t;
	for(int pos=0;pos<(*image).size;outputaux[pos++]=0x99); //clean output garbage

		  while ((cursor*2)<(*image).size) {
				t=transpose(cursor,

						(*image).width,

				(*image).height);
				put1(
					(*image).pix[cursor]
					//0xFF
					,outputaux,t>>1);
				//If there is a wasted pix don't show it
				put2(
					(*image).pix[cursor]
					//0xFF
					,outputaux,t>>1);
				//else put2(t&0xFF,outputaux,t>>1);
				cursor++;
			}
			free((*image).pix);
			(*image).pix=outputaux;
			}
			break;
		default:
		 return -1;
	}
	return i;
}

/***************************************************************\
|                         File handling                         |
\***************************************************************/

char mLoadFilePalette(char* vFile,int address, char* pal) {
	FILE *fp;
	char aux;

	if ((fp=fopen(vFile,"rb"))==NULL) {
		return 0;
	} else {
		fpos_t position=address;
		fsetpos (fp, &position);
		aux=fread (pal,SIZE_OF_PALETTE,1,fp);
		fclose(fp);
		return aux;
	}
}

int mLoadFileArray(char* vFile,char** array) {

	FILE *fp;
	int aux;

	if ((fp=fopen(vFile,"rb"))==NULL) {
		return 0;
	} else {
		fseek(fp,0,SEEK_END);
		aux=ftell(fp);
		if ( !aux || (aux>SIZE_OF_FILE) || ( ((*array=(char*)malloc(sizeof(char)*aux))==NULL) ) ) {
			printf("ohoh %d\r\n",aux);
			fclose(fp);
			return 0;
		} else {
			fseek(fp,0,SEEK_SET);
			aux=fread (*array,1,aux,fp);
			fclose(fp);
			return aux;
		}
	}
}

char mSaveRaw(char* vFile,unsigned char* output, int size) {
  FILE * pFile;
  if ((pFile = fopen (vFile , "wb"))==NULL) {
		return 0;
	}
	if (!size) {
		return 0;
	}
  fwrite (output, 1, size, pFile);
  fclose (pFile);
  return 1;
}

/***************************************************************\
|                     M A I N   P R O G R A M                   |
\***************************************************************/

int main(int argc,char* arg[]) {
	printf("Princed Graphics library (c) 2003 - Tammo Jan Dijkema, Enrique P. Calot\r\n");
	printf("PGE: Princed Graphics Extractor V1.0 Alpha 1.\r\n\r\n");

	if (argc!=2) {
		printf("Syntax:\r\n pge inputfile [>outputfile.txt]\r\n\r\nNote: inputfile must not have extension\r\n\r\n");
		return 0;
	}

	//Variables
	char*         array;
	int           fileSize;
	int           imgn  			= 1;
	int 					b   				= 0;
	char          vFileraw[]	= "     x123456789          .raw";
	char          vFilebmp[]	= "     x123456789          .bmp";
	char          vFiledat[]	= "     x123456789.dat";
	int           imgEnd      = 0;
	int           a           = 0;
  tImage        image;

	//Load file in memory
	sprintf(vFiledat,"dat\\%s.dat",arg[1]);
	mLoadFilePalette(vFiledat,0x0B, image.pal);
	fileSize=mLoadFileArray(vFiledat,&array);
	printf("Opening: %s. Size: %d bytes.\n",vFiledat,fileSize);

	while (a<(fileSize-100)) {
		//Verify img header
		//send into a function
		if ((!array[a+1])||(array[a+2])||(!array[a+3])||(array[a+4])||(array[a+5])||(((unsigned char)array[a+6]&0xF0)!=0xB0))	{
			a++;
			continue;
		}

		if ((imgEnd=mExpandGraphic(array,&image,a))>0) {
		  //Print fileinfo
			if (1) {
			sprintf(vFileraw,"raw\\%s.%05d.%2x.raw",arg[1],a,(unsigned char)array[a+6]);
			sprintf(vFilebmp,"bmp\\%s.%05d.%2x.bmp",arg[1],a,(unsigned char)array[a+6]);
			mSaveRaw(vFileraw,image.pix,(image.size+1)/2);
			mWriteBitMap(image,vFilebmp);
			}
			printf("Img: %3d. Checksum: %3d (%02X). Addr: %5d to %5d. w=%3d. h=%3d. Format: %0X. Size: %3d. Diff: %3d. File: %s.\n",imgn++,(unsigned char)array[a],(unsigned char)array[a],a,imgEnd,image.width,image.height,(unsigned char)array[a+6],(imgEnd-a),(a-b),vFilebmp);
			b=imgEnd;
			free(image.pix);
		} else {
			printf("Img: %2d. Checksum: %3d (%02X). Addr: %5d. Unknown compression format %0X. Size: %dx%d (%d). Starting with %02X (%d)\n",imgn++,(unsigned char)array[a],(unsigned char)array[a],a,(unsigned char)array[a+6],(unsigned char)array[a+1],(unsigned char)array[a+3],(unsigned char)array[a+1]*(unsigned char)array[a+3],(unsigned char)array[a+7],array[a+7]);
		}

		if (imgEnd=a) {
			a++;
		}

	}
	if (fileSize) free(array);
	return 1;
}
